/*****************************************************************************
	 malloctest.c

    Z-World, Inc, 2004

    This program demonstrates the sample malloc dynamic xmemory
    allocation system. It can work with a heap in extended
    memory or root. See comments in MALLOC.LIB.

    It gives a visual representation of the heap structure
    and usage, and illustrates how a heap can get fragmented
    and/or exhausted.

    The first part of the test mallocs uniformed sized objects
    until the heap is all used, then frees them in random order.

    The second part randomly mallocs and frees blocks which can
    be constant sized or randomly varied (see comments with msize
    below).

    This is a good test program if making modifications to
    malloc.lib
*****************************************************************************/

// Uncomment one type of heap only
#define XMEM_HEAP
//#define ROOT_HEAP

/**************************
 Array or xmem size to be used as a heap 8k bytes
 (need  two pad bytes for xmem heap). This program
 is  designed to work with this size heap only -
 DO NOT CHANGE THE SIZE!
**************************///
#define HEAP_SIZE 0x1FFE  // DO NOT CHANGE!

// Above macros needed before #use'ing MALLOC.LIB
#use "MALLOC.LIB"
#use "RAND.LIB"
#use "XMALLOC.LIB"

/*****************************************************
 This function is used to determine the size parameter
 for malloc in the second test portion of the program

  Uncomment only one return statement at a time
  don't return less than 4, as the test program
  is designed to work with a limited number of
  malloc'ed pointers, and the granualarity of the
  display is 4 bytes in any case.
******************************************************/
msize()
{
   // The heap should not cause a problem with this
   // small constant size (for THIS program!)
      return 8;

	// Uncomment this return to demonstrate sporadic
	// heap fragmentation
	//	return ranRange(13,31);

	// Uncomment this return to demonstrate severe
	// heap fragmentation
	//   return ranRange(13,49);

	// Uncomment this return to demonstrate quick heap
	// exhaustion
	//return 64;
}

unsigned topheap, botheap;
unsigned long failedMallocs;
struct PT {       // pointer table to remember
     char *ptr;   // malloc'ed pointers
     unsigned Size;
   };

root main()
{
   unsigned long paddr;
	static char *pt, rstr[0x256];
   unsigned int indx, indx2, i, j, Size;
   static struct PT p[0x200];
   struct free__list IE;

  	rand16_init(7 & (int)MS_TIMER);  // initialize rand

   //***** Initialize heap
#ifdef XMEM_HEAP
	// initialized a 4k aligned xmem location  for heap
   xheapInit(HEAP_SIZE,4);
   paddr = xmalloc4kalign(HEAP_SIZE);
#endif
   heapInit(paddr);

   IE.Size = alloc__h->Size;  // save initial entry
   IE.next = alloc__h->next;  //  to compare later

	// heap allocation starts at the top
	topheap = (int)(pt=malloc(1));
   botheap = topheap-HEAP_SIZE+1;
   free(pt);

   drawHeap();

	// initializations
   failedMallocs = 0;
   for(i=0;i<0x200;i++)
      p[i].ptr=NULL;
   i=0;
   gotoxy(1,40);
   printf("Hit any key to pause    ");

   gotoxy(0,42);
   printf(" Free Block List:");

   //***** Allocate all available 100 byte blocks
   while(1)
   {
		checkKey(); // If key is hit, busy wait until next key hit.

      // If malloc returns zero, the heap is exhausted
      if(!(p[i].ptr = malloc(100))) break;

      outputblock((unsigned)p[i].ptr-botheap,100,41);
  	  	outputStats();
  	   outputFList();
     	i++;
	}

  	outputStats();

	// Initialize random deal string
	for(j=0;j<i;j++)
   	rstr[j]=j+1;
   rstr[j]=0;

	//***** Free the 100 byte blocks in random order
	while(1)
   {
		checkKey(); // If key is hit, busy wait until next key hit.

   	if(!(indx = dealRandom(rstr))) break;
	   outputblock((unsigned)p[indx-1].ptr-botheap,100,42);
  		free(p[indx-1].ptr);
    	outputStats();
  	   outputFList();
	}
   gotoxy(0,42);
   printf("                      ");
   gotoxy(0,43);
   printf("                      ");

   //***** Every thing malloced was freed, so heap should
   //*****  be in it's initial state.
   if(IE.Size !=  alloc__h->Size ||
      IE.next !=  alloc__h->next
      )
   {
	   gotoxy(1,40);
   	printf("ERROR! Incorrectly freed memory");
      while(1);
   }

	// Initialize for second test
   for(i=0;i<0x200;i++)
   {
      p[i].ptr=NULL;
	}
   Size = msize();

   i = 0xffffu;

   while(1)
   {
		checkKey(); // If key is hit, busy wait until next key hit

   	costate{

         indx2 =  i%0x200; //select a random block to malloc
      	if(p[indx2].ptr==NULL)
         {
            // Allocate a block if the indexed structure is not in use.
            waitfor(     		// uses lazy evaluation
                ++failedMallocs &&  // always  true
            	(p[indx2].ptr = (char*)malloc(Size)) &&  //true if malloc works
					(p[indx2].Size = Size) && // always true if executed
               (Size = msize())        // always true if executed
               );

				// This runs if and only if  malloc worked last try
				outputblock((unsigned)p[indx2].ptr-botheap,p[indx2].Size,41);
		      failedMallocs--;
            i++;
         }
      }
      indx =  ranRange(2,0x400)/2-1; //select a random block to free
      if(p[indx].ptr)  // if pointer table entry in use
      {
        	free(p[indx].ptr);
        	outputblock((unsigned)p[indx].ptr-botheap,p[indx].Size, 42);
        	p[indx].ptr=NULL;
     		outputStats();
      }
   }
}

/*****************************************************
  Draw heap on screen -
  The escape sequence \x1b[XXm , where XX is the
  attribute byte described in the sample program
  enum.c is used throughout this program to set
  set output colors. The sequence used in gotoxy()
  is used directly for efficiently placing the
  cursor also since extra printf calls slow things
  down. Although efficient, it makes some cryptic
  looking printf strings.
*****************************************************/
nodebug
drawHeap()
{
   int i;

	outputStats();
   box(1, 1, 65, 33);
   gotoxy(42,35);
	printf("\x1b[41m  \x1b[0m = malloc'ed space");
   gotoxy(42,37);
	printf("\x1b[42m  \x1b[0m = free space");
	for(i=0;i<32;i++)
   {  // print addresses down side
		printf("\x1B=%c%c%04x",68+0x20,0x22+i,botheap+i*0x100);
   }
  	printf(
    "\x1B=$CNumber of free blocks ="  \
    "\x1B=$D     Total free space ="  \
    "\x1B=$E     # Failed mallocs ="  );
}

nodebug
checkKey()
{
	if(kbhit())
   {
	   gotoxy(1,40);
 		printf("Hit any key to continue");
     	getchar(); while(!kbhit()); getchar();
	   gotoxy(1,40);
	   printf("Hit any key to pause    ");
   }
}

// output statistics
nodebug
root outputStats()
 {
	struct free__list *p;
   unsigned int count, Tsize;
   unsigned row, col;

   ENTER_SET_XPC;    // needed in case heap in xmem

   count = 0;
   Tsize = 0;
   p = alloc__h;    // initial free list entry is always in alloc__h

	while(p)
   {
    	Tsize += p->Size;
      p = p->next;
      count += 1;
	}
	RESTORE_XPC_EXIT;	// needed in case heap in xmem

	// Use escape sequences to position cursor
   // and send everything in one shot.
	printf("\x1B=<C%04x\x1B=<D%04x\x1B=<E%08lx",
	                         count,
                            Tsize,
                            failedMallocs      );
}

nodebug
root outputblock(unsigned start, unsigned Size, int color)
{
  unsigned int i, sz, row, col;
   static char str[512],str1[512];

	sprintf(str,"\x1b[%dm",color);
   for(i=start;i<start+Size;i = i+1)
   {
    	col = (i/4) % 64;
 		row = ((i)/8) / 32 ;
	   sprintf (str1,"\x1B=%c%c ",0x22+col,0x22+row );
	   strcat(str,str1);
	}
   strcat(str,"\x1b[0m");
  	if((sz=strlen(str))<129)
   {
  		printf("%s",str);
   }
   else
   {
      for(i=0; i < (sz/127+1); i++)
      {
			memcpy(str1,(char*)((unsigned)str+i*127),128);
   	   str1[127]=0;
   		printf("%s",str1);
	   }
	}
}

// Position Cursor
nodebug
gotoxy (int x, int y)
{
   x += 0x20;
   y += 0x20;
   printf ("\x1B=%c%c",x,y );
}

// Print a green bordered box with upper right corner
// at (starCol,startRow).
nodebug
box(int startCol,int startRow, int width, int height)
{
	int i;
	char hor_line[70], greenrow[70];

	// Define horizontal border and green row
	for( i = 0 ; i < width-1 ; i++)
	{
		hor_line[i] = 0xC4;
      greenrow[i] = ' ';
	}
	hor_line[i] = greenrow[i] = 0;

	// Top of table border
   printf ( "\x1B=%c%c%c%s%c",
	   0x20+startCol, 0x20+startRow, 0xDA,  // upper left corner
      hor_line,                            // top middle
	   0xBF                                 // upper right corner
     );

	// Table middle including borders
	for( i = 1 ; i < height ; i++)
	   printf ( "\x1B=%c%c\xB3\x1b[42m%s\x1b[0m\xB3",
      	0x20+startCol, 0x20+startRow+i, greenrow );

	// Bottom of table border
   printf ( "\x1B=%c%c%c%s%c",
		0x20+startCol, 0x20+startRow+height, 0xC0, // lower left corner
      hor_line,                                  // bottom middle
      0xD9	                                     // lower right corner
     );
}

nodebug
ranRange(int min, int max)
{
	// return a random number between min & max inclusive
	return min + (rand16() % (1+max-min));
}

/*****************************************************
	Takes a NULL terminated string of arbitrary,
   non-zero characters, removes one character
   from a random index location, shortens the
   string by one, and returns the value removed.
   Returns zero for empty string.
*****************************************************/
nodebug
int dealRandom(char* str)
{
   int len,indx;
   char *ptr, retval;

   retval = 0;
   len = strlen(str);
	if(len)    // just return 0 if NULL string
   {
	   indx = rand16() % len;
   	retval = str[indx];
	   str[indx] = 0;
   	ptr = &str[indx+1];
	   strcat(str,ptr);
	}
   return retval;
}

nodebug
root outputFList()
{
   // prints the whole list of free heap blocks
   // in 3 column by ten row format

	struct free__list *p;
   unsigned int i;
   ENTER_SET_XPC;    // needed in case heap in xmem

   p = alloc__h;
	i = 0;
	while(p)
   {
      if(p->next)
      {
			printf("\x1B=%c%cA=%04x,S=%04x,N=%04x",
         	0x21+((i/10)%3)*22,0x20+43+(i%10),  // row column
            p, p->Size, p->next                 // data
          );
		}
      else
      {  // print last entry in red (ugly but saves extra printf calls
         //  to set cursor location and color)
      	printf("\x1B=%c%c\x1b[31mA=%04x,S=%04x,N=%04x\x1b[0m\x1B=%c%c" \
                "                    ",                // blank out next entry
         	0x21+((i/10)%3)*22,0x20+43+(i%10),         // this entry row,column
            p, p->Size, p->next,                       // data
            0x21+(((i+1)/10)%3)*22,0x20+43+((i+1)%10)  // next entry row,column
          );
		}
      p = p->next;
      i++;
	}
	RESTORE_XPC_EXIT;  // needed in case heap in xmem
}