/******************************************************************

 sin_table.c  - Example of using a table in xmem as a look-up
 table for a calculation.

 Z-World, Inc, 2004

 Instead of calculating a value on the fly, we use a table to simply
 look the value up. The SIN(x) table in this case is 8192 bytes, and
 only  works on first quadrant values of [0-PI/2] radians. This method
 takes about half the time of the intrinsic sin() function in Dynamic
 C's  libraries. More complex functions would save more time. The SIN(x)
 table also gives COS(x) using the indentity cos(x) = sin(PI/2-x)

 The total table size is 32768 bytes. It can be shown that the maximum
 error at this granularity is given by (PI/2)/8192 or about 0.0002. More
 complex implementations could use rounding or linear interpolation to get
 more accuracy. If less accuracy is required a smaller table can be used.

 NOTE: The intrinsic sin() function uses fixed point and is quite fast,
 and it is not limited to the first quadrant, but the performance is
 a bit slower when the input is not in the first quadrant. This method
 is presented as an alternative method to using a look-up table in root
 for more complex functions. Fixed point methods (some of which are used
 in Dynamic C's math libraries) are another alternative for floating point
 code.

   // This is the data generating program which was compiled with
   // Borland. (It won't compile with Dynamic C.)
   // The pre-created data file is included with this Tech Note package.
	#include <stdio.h>
	#include <math.h>
	#define PI ((double)3.1415926535897932384626433832795)

	void main()	{
		int i;
	   float x;
		FILE * fp;
	   fp = fopen("sindata.bin","wb");
	   for(i=0; i<8192; i++)
   	{
			x = sin((float)(i)*PI/16384.0);
   	   fwrite(&x,sizeof(float),1,fp);
	   }
   	fflush(fp);
	   fclose(fp);
	}

******************************************************************/

#ximport "sindata.bin" sin_tab

// The index is simply scaled on the interval [0,8191] the first 4
// bytes  at the physical address sin_tab are the size of the file.
// The floating point multiplication and division add a few cycles
#define SIN(x) (xgetfloat(sin_tab + 4U + 4U*(unsigned)((x)*16384.0/PI)))

// The index for COS(x) is 32764 minus the index for SIN(X)
#define COS(x) (xgetfloat(sin_tab + 32768U - 4U*(unsigned)((x)*16384.0/PI)))

main()
{
	int i;
	float x,y,z;

	printf(" x (rad.)    SIN(x)      COS(x)     |S(x)-s(x)|  |C(x)-cos(x)| \n");
	printf("-------------------------------------------------------------\n");

   for(i=0; i<=8192; i = i+ 256)
   {
		x = (float)(i)*PI/16384.0; // = (float)i*(PI/2)/8192
      y = SIN(x);
      z = COS(x);

		// display values and absolute errors measured against the
      // the intrinsic sin() and cos() functions
		printf("%9.6f   %9.6f   %9.6f   %9.6f   %9.6f \n",
      	x, y, z, fabs(y-sin(x)), fabs(z-cos(x)));
   }
}


